@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2008-2012, Parallels, Inc. All rights reserved.
+ *  Copyright (C) 2008-2014, Parallels, Inc. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,11 +16,6 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
-/* _XXX_ We should use AIO. ploopcopy cannot use cached reads and
- * has to use O_DIRECT, which introduces large read latencies.
- * AIO is necessary to transfer with maximal speed.
- */
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <errno.h>
@@ -33,11 +28,37 @@
 #include <linux/types.h>
 #include <linux/fs.h>
 #include <string.h>
+#include <pthread.h>
 
 #include "ploop.h"
 
 #define SYNC_MARK	4
 
+#define STATUS_OK	"SUCCESS\n"
+#define STATUS_FAIL	"FAILURE\n"
+#define LEN_STATUS	8
+
+/* Check what a file descriptor refers to.
+ * Return:
+ *  0 - file
+ *  1 - pipe or socket
+ * -1 - none of the above
+ */
+static int is_fd_pipe(int fd) {
+	struct stat st;
+
+	if (fstat(fd, &st))
+		return -1;
+
+	if (S_ISREG(st.st_mode))
+		return 0;
+
+	if (S_ISFIFO(st.st_mode) || S_ISSOCK(st.st_mode))
+		return 1;
+
+	return -1;
+}
+
 static int nwrite(int fd, const void *buf, int len)
 {
 	while (len) {
@@ -135,21 +156,41 @@ static int nread(int fd, void * buf, int len)
 	return -1;
 }
 
-int ploop_receive(const char *dst)
+int ploop_copy_receive(struct ploop_copy_receive_param *arg)
 {
 	int ofd, ret;
 	__u64 cluster = 0;
 	void *iobuf = NULL;
 
-	if (isatty(0) || errno == EBADF) {
-		ploop_err(errno, "Invalid input stream: must be pipelined "
-				"to a pipe or a socket");
+	if (!arg)
+		return SYSEXIT_PARAM;
+
+	if (is_fd_pipe(arg->ifd) != 1) {
+		ploop_err(errno, "Invalid input fd %d: must be "
+				"a pipe or a socket", arg->ifd);
 		return SYSEXIT_PARAM;
 	}
 
-	ofd = open(dst, O_WRONLY|O_CREAT|O_EXCL, 0600);
+	if (arg->feedback_fd >= 0 && is_fd_pipe(arg->feedback_fd) != 1) {
+		ploop_err(errno, "Invalid feedback fd %d: must be "
+				"a pipe or a socket", arg->feedback_fd);
+		return SYSEXIT_PARAM;
+	}
+
+	/* If feedback is to be send to stdout or stderr,
+	 * we have to disable logging to appropriate fd.
+	 *
+	 * As currently there's no way to disable just stderr,
+	 * so in this case we have to disable stdout as well.
+	 */
+	if (arg->feedback_fd == STDOUT_FILENO)
+		ploop_set_verbose_level(PLOOP_LOG_NOSTDOUT);
+	else if (arg->feedback_fd == STDERR_FILENO)
+		ploop_set_verbose_level(PLOOP_LOG_NOCONSOLE);
+
+	ofd = open(arg->file, O_WRONLY|O_CREAT|O_EXCL, 0600);
 	if (ofd < 0) {
-		ploop_err(errno, "Can't open %s", dst);
+		ploop_err(errno, "Can't open %s", arg->file);
 		return SYSEXIT_CREAT;
 	}
 
@@ -158,8 +199,8 @@ int ploop_receive(const char *dst)
 		int n;
 		struct xfer_desc desc;
 
-		if (nread(0, &desc, sizeof(desc)) < 0) {
-			ploop_err(0, "Error in nread(desc)");
+		if (nread(arg->ifd, &desc, sizeof(desc)) < 0) {
+			ploop_err(errno, "Error in nread(desc)");
 			ret = SYSEXIT_READ;
 			goto out;
 		}
@@ -180,20 +221,35 @@ int ploop_receive(const char *dst)
 		if (desc.size == 0)
 			break;
 
-		if (nread(0, iobuf, desc.size)) {
+		if (nread(arg->ifd, iobuf, desc.size)) {
 			ploop_err(errno, "Error in nread data");
 			ret = SYSEXIT_READ;
 			goto out;
 		}
 		if (desc.size == SYNC_MARK) {
+			int st;
 			/* ignore received data, instead do sync */
-			ploop_log(4, "%s: fdatasync()", __func__);
-			if (fdatasync(ofd)) {
+			st = fdatasync(ofd);
+			if (arg->feedback_fd >= 0) {
+				/* Tell the sending side how it went */
+				int w;
+
+				w = write(arg->feedback_fd,
+					st ? STATUS_FAIL : STATUS_OK,
+					LEN_STATUS);
+				/* check write error only if no error yet */
+				if (!st && w != LEN_STATUS) {
+					ploop_err(errno, "Error in write(%d)",
+							arg->feedback_fd);
+					ret = SYSEXIT_WRITE;
+					goto out;
+				}
+			}
+			if (st) {
 				ploop_err(errno, "Error in fdatasync()");
 				ret = SYSEXIT_WRITE;
 				goto out;
 			}
-			ploop_log(4, "%s: fdatasync() complete", __func__);
 			continue;
 		}
 		n = pwrite(ofd, iobuf, desc.size, desc.pos);
@@ -222,12 +278,21 @@ out:
 			ret = SYSEXIT_WRITE;
 	}
 	if (ret)
-		unlink(dst);
+		unlink(arg->file);
 	free(iobuf);
 
 	return ret;
 }
 
+/* Deprecated, use ploop_copy_receive() instead */
+int ploop_receive(const char *dst) {
+	struct ploop_copy_receive_param r = {
+		.file = dst,
+	};
+
+	return ploop_copy_receive(&r);
+}
+
 static int get_image_info(const char *device, char **send_from_p,
 		char **format_p, int *blocksize)
 {
@@ -277,12 +342,10 @@ static int run_cmd(const char *cmd)
 
 static int open_mount_point(const char *device)
 {
-	int ret;
 	int fd;
 	char mnt[PATH_MAX];
 
-	ret = ploop_get_mnt_by_dev(device, mnt, sizeof(mnt));
-	if (ret == -1) {
+	if (ploop_get_mnt_by_dev(device, mnt, sizeof(mnt))) {
 		ploop_err(0, "Can't find mount point for %s", device);
 		return -1;
 	}
@@ -296,8 +359,85 @@ static int open_mount_point(const char *device)
 	return fd;
 }
 
-int ploop_send(const char *device, int ofd, const char *flush_cmd,
-		int is_pipe)
+/* If you want to see critical ploop_send() timings:
+ *
+ * 1) Compile with DEBUG_TIMES defined
+ * 2) Use "ploop -vvvv copy" to set verbosity to (at least) 4
+ * 3) Use either -d FILE or -o OUTFD (to keep stdout free for logging)
+ */
+#ifdef DEBUG_TIMES
+#define TS(...) ploop_log(4, "psend: " __VA_ARGS__)
+#else
+#define TS(...)
+#endif
+
+struct send_data {
+	int fd;
+	void *buf;
+	int len;
+	off_t pos;
+	int is_pipe;
+	int ret;
+	int err_no;
+	pthread_mutex_t mutex;
+	pthread_cond_t cond;
+	int has_data;
+};
+
+static void *send_thread(void *data) {
+	struct send_data *sd = data;
+	int done;
+
+	do {
+		pthread_mutex_lock(&sd->mutex);
+		if (!sd->has_data) {
+			pthread_cond_wait(&sd->cond, &sd->mutex);
+		}
+		sd->ret = send_buf(sd->fd, sd->buf,
+				sd->len, sd->pos, sd->is_pipe);
+		if (sd->ret)
+			sd->err_no = errno;
+		done = (sd->len == 0 && sd->pos == 0);
+		sd->has_data = 0;
+		pthread_mutex_unlock(&sd->mutex);
+	} while (!done);
+
+	return NULL;
+}
+
+#define async_send(length, position)			\
+do {							\
+	pthread_mutex_lock(&sd.mutex);			\
+	if (sd.ret) {					\
+		ploop_err(sd.err_no, "write error");	\
+		ret = sd.ret;				\
+		goto done;				\
+	}						\
+	sd.buf = iobuf[i];				\
+	sd.len = (length);				\
+	sd.pos = (position);				\
+	sd.has_data = 1;				\
+	pthread_mutex_unlock(&sd.mutex);		\
+	pthread_cond_signal(&sd.cond);			\
+} while (0)
+
+#define do_pread(length, position)			\
+({							\
+	int __ret;					\
+							\
+	i = !i;						\
+	__ret = idelta.fops->pread(idelta.fd, iobuf[i],	\
+			(length), (position));		\
+	if (__ret < 0) {				\
+		ploop_err(errno, "Error from read");	\
+		ret = SYSEXIT_READ;			\
+		goto done;				\
+	}						\
+							\
+	__ret;						\
+})
+
+int ploop_copy_send(struct ploop_copy_send_param *arg)
 {
 	struct delta idelta = { .fd = -1 };
 	int tracker_on = 0;
@@ -307,7 +447,7 @@ int ploop_send(const char *device, int ofd, const char *flush_cmd,
 	int ret = 0;
 	char *send_from = NULL;
 	char *format = NULL;
-	void *iobuf = NULL;
+	void *iobuf[2] = {};
 	int blocksize;
 	__u64 cluster;
 	__u64 pos;
@@ -317,36 +457,66 @@ int ploop_send(const char *device, int ofd, const char *flush_cmd,
 	__u64 xferred;
 	int iter;
 	struct ploop_track_extent e;
+	int i;
+	pthread_t send_th = 0;
+	struct send_data sd = {
+		.mutex = PTHREAD_MUTEX_INITIALIZER,
+		.cond = PTHREAD_COND_INITIALIZER,
+	};
 
-	// Do not print anything on stdout, since we use it to send delta
-	if (is_pipe && ofd == STDOUT_FILENO)
+	if (!arg)
+		return SYSEXIT_PARAM;
+
+	sd.fd = arg->ofd;
+	sd.is_pipe = is_fd_pipe(arg->ofd);
+	if (sd.is_pipe < 0) {
+		ploop_err(0, "Invalid output fd %d: must be a file, "
+				"a pipe or a socket", arg->ofd);
+		return SYSEXIT_PARAM;
+	}
+
+	if (arg->feedback_fd >= 0 && is_fd_pipe(arg->feedback_fd) != 1) {
+		ploop_err(errno, "Invalid feedback fd %d: must be "
+				"a pipe or a socket", arg->feedback_fd);
+		return SYSEXIT_PARAM;
+	}
+
+	/* If data is to be send to stdout or stderr,
+	 * we have to disable logging to appropriate fd.
+	 *
+	 * As currently there's no way to disable just stderr,
+	 * so in this case we have to disable stdout as well.
+	 */
+	if (arg->ofd == STDOUT_FILENO)
 		ploop_set_verbose_level(PLOOP_LOG_NOSTDOUT);
+	else if (arg->ofd == STDERR_FILENO)
+		ploop_set_verbose_level(PLOOP_LOG_NOCONSOLE);
 
-	devfd = open(device, O_RDONLY);
+	devfd = open(arg->device, O_RDONLY);
 	if (devfd < 0) {
-		ploop_err(errno, "Can't open device %s", device);
+		ploop_err(errno, "Can't open device %s", arg->device);
 		ret = SYSEXIT_DEVICE;
 		goto done;
 	}
 
-	mntfd = open_mount_point(device);
+	mntfd = open_mount_point(arg->device);
 	if (mntfd < 0) {
 		/* Error is printed by open_mount_point() */
 		ret = SYSEXIT_OPEN;
 		goto done;
 	}
 
-	ret = get_image_info(device, &send_from, &format, &blocksize);
+	ret = get_image_info(arg->device, &send_from, &format, &blocksize);
 	if (ret)
 		goto done;
 	cluster = S2B(blocksize);
 
-	if (p_memalign(&iobuf, 4096, cluster)) {
-		ret = SYSEXIT_MALLOC;
-		goto done;
-	}
+	ret = SYSEXIT_MALLOC;
+	for (i = 0; i < 2; i++)
+		if (p_memalign(&iobuf[i], 4096, cluster))
+			goto done;
 
-	ret = ploop_complete_running_operation(device);
+	ret = ploop_complete_running_operation(arg->device);
 	if (ret)
 		goto done;
 
@@ -360,6 +530,13 @@ int ploop_send(const char *device, int ofd, const char *flush_cmd,
 		goto done;
 	}
 
+	ret = pthread_create(&send_th, NULL, send_thread, &sd);
+	if (ret) {
+		ploop_err(ret, "Can't create send thread");
+		ret = SYSEXIT_SYS;
+		goto done;
+	}
+
 	ploop_log(-1, "Sending %s", send_from);
 
 	trackend = e.end;
@@ -371,21 +548,11 @@ int ploop_send(const char *device, int ofd, const char *flush_cmd,
 		if (ret)
 			goto done;
 
-		n = idelta.fops->pread(idelta.fd, iobuf, cluster, pos);
-		if (n < 0) {
-			ploop_err(errno, "pread");
-			ret = SYSEXIT_READ;
-			goto done;
-		}
-		if (n == 0)
+		n = do_pread(cluster, pos);
+		if (n == 0) /* EOF */
 			break;
 
-		ret = send_buf(ofd, iobuf, n, pos, is_pipe);
-		if (ret) {
-			ploop_err(errno, "write");
-			goto done;
-		}
-
+		async_send(n, pos);
 		pos += n;
 	}
 	/* First copy done */
@@ -424,22 +591,13 @@ int ploop_send(const char *device, int ofd, const char *flush_cmd,
 						goto done;
 					}
 				}
-				n = idelta.fops->pread(idelta.fd, iobuf, copy, pos);
-				if (n < 0) {
-					ploop_err(errno, "read2");
-					ret = SYSEXIT_READ;
-					goto done;
-				}
+				n = do_pread(copy, pos);
 				if (n == 0) {
-					ploop_err(0, "unexpected EOF");
+					ploop_err(0, "Unexpected EOF");
 					ret = SYSEXIT_READ;
 					goto done;
 				}
-				ret = send_buf(ofd, iobuf, n, pos, is_pipe);
-				if (ret) {
-					ploop_err(errno, "write2");
-					goto done;
-				}
+				async_send(n, pos);
 				pos += n;
 			}
 		} else {
@@ -470,31 +628,68 @@ int ploop_send(const char *device, int ofd, const char *flush_cmd,
 	 * this as "sync me" command, while the old one just writes those
 	 * bytes which is useless but harmless.
 	 */
-	ret = idelta.fops->pread(idelta.fd, iobuf, 4096, 0);
-	if (ret != 4096) {
-		ploop_err(errno, "pread");
-		ret = SYSEXIT_READ;
-		goto done;
-	}
-	ret = send_buf(ofd, iobuf, SYNC_MARK, 0, is_pipe);
-	if (ret) {
-		ploop_err(errno, "write");
-		goto done;
-	}
-	/* Now we should wait for the other side to finish syncing
-	 * before freezing the container. This is done in order to
-	 * optimize CT frozen time. Unfortunately the protocol is
-	 * one-way so there is no way to receive anything from the
-	 * other side. As ugly as it is, let's just sleep for some time.
-	 */
-	sleep(5);
+	if (sd.is_pipe) {
+		char buf[LEN_STATUS + 1] = {};
 
+		ret = do_pread(4096, 0);
+		if (ret < SYNC_MARK) {
+			ploop_err(errno, "Short read");
+			ret = SYSEXIT_READ;
+			goto done;
+		}
+		TS("SEND 0 %d (sync)", SYNC_MARK);
+		async_send(SYNC_MARK, 0);
+
+		/* Now we should wait for the other side to finish syncing
+		 * before freezing the container, to optimize CT frozen time.
+		 */
+		if (arg->feedback_fd < 0) {
+			/* No descriptor to receive a response back is given.
+			 * As ugly as it looks, let's just sleep for some time
+			 * hoping the other side will finish sync.
+			 */
+			TS("SLEEP 5");
+			sleep(5);
+			goto sync_done;
+		}
+
+		/* Wait for feedback from the receiving side */
+
+		/* FIXME: use select/poll with a timeout */
+		if (read(arg->feedback_fd, buf, LEN_STATUS)
+				!= LEN_STATUS) {
+			ploop_err(errno, "Can't read feedback");
+			ret = SYSEXIT_PROTOCOL;
+			goto done;
+		}
+
+		if (strncmp(buf, STATUS_OK, LEN_STATUS) == 0) {
+			goto sync_done;
+		}
+		else if (strncmp(buf, STATUS_FAIL, LEN_STATUS) == 0) {
+			ploop_err(0, "Remote side reported sync failure");
+			ret = SYSEXIT_FSYNC;
+			goto done;
+		}
+		else {
+			ploop_err(0, "Got back feedback: %s", buf);
+			ret = SYSEXIT_PROTOCOL;
+			goto done;
+		}
+	} else {
+		/* Writing to local file */
+		fdatasync(arg->ofd);
+	}
+
+sync_done:
 	/* Freeze the container */
-	ret = run_cmd(flush_cmd);
+	TS("FLUSH");
+	ret = run_cmd(arg->flush_cmd);
 	if (ret)
 		goto done;
 
 	/* Sync fs */
+	TS("SYNCFS");
 	if (sys_syncfs(mntfd)) {
 		ploop_err(errno, "syncfs() failed");
 		ret = SYSEXIT_FSYNC;
@@ -502,11 +697,13 @@ int ploop_send(const char *device, int ofd, const char *flush_cmd,
 	}
 
 	/* Flush journal and freeze fs (this also clears the fs dirty bit) */
+	TS("FIFREEZE");
 	ret = ioctl_device(mntfd, FIFREEZE, 0);
 	if (ret)
 		goto done;
 	fs_frozen = 1;
 
+	TS("IOC_SYNC");
 	ret = ioctl_device(devfd, PLOOP_IOC_SYNC, 0);
 	if (ret)
 		goto done;
@@ -542,22 +739,15 @@ int ploop_send(const char *device, int ofd, const char *flush_cmd,
 					if (ret)
 						goto done;
 				}
-				n = idelta.fops->pread(idelta.fd, iobuf, copy, pos);
-				if (n < 0) {
-					ploop_err(errno, "read3");
-					ret = SYSEXIT_READ;
-					goto done;
-				}
+				TS("READ %llu %d", pos, copy);
+				n = do_pread(copy, pos);
 				if (n == 0) {
-					ploop_err(0, "unexpected EOF3");
+					ploop_err(0, "Unexpected EOF");
 					ret = SYSEXIT_READ;
 					goto done;
 				}
-				ret = send_buf(ofd, iobuf, n, pos, is_pipe);
-				if (ret) {
-					ploop_err(errno, "write3");
-					goto done;
-				}
+				TS("SEND %llu %d", pos, n);
+				async_send(n, pos);
 				pos += n;
 			}
 		} else {
@@ -579,41 +769,43 @@ int ploop_send(const char *device, int ofd, const char *flush_cmd,
 	/* Must clear dirty flag on ploop1 image. */
 	if (strcmp(format, "ploop1") == 0) {
 		int n;
-		struct ploop_pvd_header *vh = (void*)iobuf;
+		struct ploop_pvd_header *vh;
 
-		n = idelta.fops->pread(idelta.fd, iobuf, 4096, 0);
-		if (n != 4096) {
-			ploop_err(errno, "Error reading 1st sector of %s", send_from);
+		TS("READ 0 4096");
+		n = do_pread(4096, 0);
+		if (n < SECTOR_SIZE) {
+			ploop_err(errno, "Short read");
 			ret = SYSEXIT_READ;
 			goto done;
 		}
 
+		vh = iobuf[i];
 		vh->m_DiskInUse = 0;
 
-		ret = send_buf(ofd, vh, SECTOR_SIZE, 0, is_pipe);
-		if (ret) {
-			ploop_err(errno, "write3");
-			goto done;
-		}
+		TS("SEND 0 %d (1st sector)", SECTOR_SIZE);
+		async_send(SECTOR_SIZE, 0);
 	}
 
+	TS("IOCTL TRACK_STOP");
 	ret = ioctl(devfd, PLOOP_IOC_TRACK_STOP, 0);
 	if (ret)
 		goto done;
 	tracker_on = 0;
 
-	ret = send_buf(ofd, iobuf, 0, 0, is_pipe);
-	if (ret) {
-		ploop_err(errno, "write4");
-		goto done;
-	}
+	TS("SEND 0 0 (close)");
+	async_send(0, 0);
+	pthread_join(send_th, NULL);
+	send_th = 0;
 
 done:
+	if (send_th)
+		pthread_cancel(send_th);
 	if (fs_frozen)
 		(void)ioctl_device(mntfd, FITHAW, 0);
 	if (tracker_on)
 		(void)ioctl_device(devfd, PLOOP_IOC_TRACK_ABORT, 0);
-	free(iobuf);
+	free(iobuf[0]);
+	free(iobuf[1]);
 	if (devfd >=0)
 		close(devfd);
 	if (mntfd >=0)
@@ -622,5 +814,21 @@ done:
 	if (idelta.fd >= 0)
 		close_delta(&idelta);
 
+	TS("DONE");
 	return ret;
 }
+#undef do_pread
+#undef async_send
+
+/* Deprecated, please use ploop_copy_send() instead */
+int ploop_send(const char *device, int ofd, const char *flush_cmd,
+		int is_pipe)
+{
+	struct ploop_copy_send_param s = {
+		.device		= device,
+		.ofd		= ofd,
+		.flush_cmd	= flush_cmd,
+	};
+
+	return ploop_copy_send(&s);
+}
