@@ -34,6 +34,7 @@
 #include <sys/syscall.h>
 #include <mntent.h>
 #include <ext2fs/ext2_fs.h>
+#include <stdint.h>
 
 #include "ploop.h"
 #include "cleanup.h"
@@ -247,25 +248,55 @@ static int default_fmt_version(void)
 	return ploop_is_large_disk_supported() ? PLOOP_FMT_V2 : PLOOP_FMT_V1;
 }
 
+static int get_max_ploop_size(int version, unsigned int blocksize, __u64 *max)
+{
+	switch(version) {
+	case PLOOP_FMT_V1:
+		*max = (__u32)-1;
+		break;
+	case PLOOP_FMT_V2:
+		*max = 0xffffffffUL * blocksize;
+		break;
+	case PLOOP_FMT_UNDEFINED:
+		*max = UINT64_MAX;
+		break;
+	default:
+		ploop_err(0, "Unknown ploop image version: %d", version);
+		return -1;
+	}
+	return 0;
+}
+
+/* Returns maximum ploop size
+ *
+ * blocksize	- blocksize in sectors, defaults to
+ *		  2048 sectors (1Mb block) if unset
+ *
+ * Note: default ploop format version is used
+ */
+int ploop_get_max_size(unsigned int blocksize, unsigned long long *max)
+{
+	blocksize = blocksize ?  blocksize : (1 << PLOOP1_DEF_CLUSTER_LOG);
+
+	if (get_max_ploop_size(default_fmt_version(), blocksize, max))
+		return SYSEXIT_PARAM;
+
+	if (*max > B2S(PLOOP_MAX_FS_SIZE))
+		*max = B2S(PLOOP_MAX_FS_SIZE);
+
+	return 0;
+}
+
 static int do_check_size(unsigned long long sectors, __u32 blocksize, int version,
 		__u64 max_fs_size)
 {
 	__u64 max;
 
-	switch(version) {
-	case PLOOP_FMT_V1:
-		max = (__u32)-1;
-		break;
-	case PLOOP_FMT_V2:
-		max = 0xffffffffUL * blocksize;
-		break;
-	case PLOOP_FMT_UNDEFINED:
-		/* RAW */
+	if (version == PLOOP_FMT_UNDEFINED)
 		return 0;
-	default:
-		ploop_err(0, "Unknown ploop image version: %d", version);
+
+	if (get_max_ploop_size(version, blocksize, &max))
 		return -1;
-	}
 
 	if (max_fs_size != 0 && max > B2S(max_fs_size))
 		max = B2S(max_fs_size);
@@ -534,21 +565,17 @@ static void fill_diskdescriptor(struct ploop_pvd_header *vh, struct ploop_disk_i
 	di->sectors = vh->m_Sectors;
 }
 
-static int create_image(struct ploop_disk_images_data *di,
-		const char *file, __u32 blocksize, off_t size_sec, int mode,
+static int create_image(const char *file, __u32 blocksize, off_t size_sec, int mode,
 		int version)
 {
 	int fd = -1;
-	int ret;
-	struct ploop_pvd_header vh = {};
-	char fname[PATH_MAX];
-	char ddxml[PATH_MAX];
 
 	if (size_sec == 0) {
 		ploop_err(0, "Incorrect block device size specified: "
 				"%lu sectors", (long)size_sec);
 		return SYSEXIT_PARAM;
 	}
+
 	if (file == NULL) {
 		ploop_err(0, "Image file name not specified");
 		return SYSEXIT_PARAM;
@@ -559,16 +586,6 @@ static int create_image(struct ploop_disk_images_data *di,
 		return SYSEXIT_PARAM;
 	}
 
-	get_disk_descriptor_fname_by_image(file, ddxml, sizeof(ddxml));
-	if (access(ddxml, F_OK) == 0) {
-		ploop_err(EEXIST, "Can't create %s", ddxml);
-		return SYSEXIT_PARAM;
-	}
-
-	di->size = size_sec;
-	di->mode = mode;
-
-	ret = SYSEXIT_CREAT;
 	if (mode == PLOOP_RAW_MODE)
 		fd = create_raw_delta(file, size_sec);
 	else if (mode == PLOOP_EXPANDED_MODE)
@@ -576,28 +593,11 @@ static int create_image(struct ploop_disk_images_data *di,
 	else if (mode == PLOOP_EXPANDED_PREALLOCATED_MODE)
 		fd = create_empty_preallocated_delta(file, blocksize, size_sec, version);
 	if (fd < 0)
-		goto err;
+		return SYSEXIT_CREAT;
+
 	close(fd);
 
-	generate_pvd_header(&vh, size_sec, blocksize, version);
-	fill_diskdescriptor(&vh, di);
-
-	if (realpath(file, fname) == NULL) {
-		ploop_err(errno, "failed realpath(%s)", file);
-		goto err;
-	}
-
-	if (ploop_di_add_image(di, fname, TOPDELTA_UUID, NONE_UUID)) {
-		ret = SYSEXIT_MALLOC;
-		goto err;
-	}
-
-	ret = ploop_store_diskdescriptor(ddxml, di);
-err:
-	if (ret)
-		unlink(file);
-
-	return ret;
+	return 0;
 }
 
 static int create_balloon_file(struct ploop_disk_images_data *di,
@@ -703,25 +703,36 @@ static int ploop_drop_image(struct ploop_disk_images_data *di)
 	return 0;
 }
 
-int ploop_create_image(struct ploop_create_param *param)
+static int init_dd(struct ploop_disk_images_data **di,
+		const char *ddxml, struct ploop_create_param *param)
 {
-	struct ploop_disk_images_data *di;
-	int ret;
+	struct ploop_pvd_header vh = {};
+	int fmt_version;
 	__u32 blocksize;
-	int version = param->fmt_version;
 
-	if (!is_fmt_version_valid(version)) {
-		ploop_err(0, "Unknown ploop image version: %d",
-				version);
+	if (access(ddxml, F_OK) == 0) {
+		ploop_err(EEXIST, "Can't create %s", ddxml);
 		return SYSEXIT_PARAM;
 	}
-	if (version == PLOOP_FMT_UNDEFINED)
-		version = default_fmt_version();
+
+	fmt_version = param->fmt_version == PLOOP_FMT_UNDEFINED ?
+		default_fmt_version() : param->fmt_version;
 
 	blocksize = param->blocksize ?
-			param->blocksize : (1 << PLOOP1_DEF_CLUSTER_LOG);
+		param->blocksize : (1 << PLOOP1_DEF_CLUSTER_LOG);
 
-	if (check_size(param->size, blocksize, version))
+	if (param->image == NULL) {
+		ploop_err(0, "Image file name not specified");
+		return SYSEXIT_PARAM;
+	}
+
+	if (!is_fmt_version_valid(fmt_version)) {
+		ploop_err(0, "Unknown ploop image version: %d",
+				fmt_version);
+		return SYSEXIT_PARAM;
+	}
+
+	if (check_size(param->size, blocksize, fmt_version))
 		return SYSEXIT_PARAM;
 
 	if (!is_valid_blocksize(blocksize)) {
@@ -730,23 +741,98 @@ int ploop_create_image(struct ploop_create_param *param)
 		return SYSEXIT_PARAM;
 	}
 
-	di = alloc_diskdescriptor();
-	if (di == NULL)
+	*di = alloc_diskdescriptor();
+	if (*di == NULL)
 		return SYSEXIT_MALLOC;
-	di->blocksize = blocksize;
-	ret = create_image(di, param->image, di->blocksize,
-			round_bdsize(param->size, di->blocksize, version),
-			param->mode, version);
+
+	(*di)->size = round_bdsize(param->size, blocksize, fmt_version);
+	(*di)->blocksize = blocksize;
+	(*di)->mode = param->mode;
+
+	generate_pvd_header(&vh, (*di)->size, blocksize, fmt_version);
+	fill_diskdescriptor(&vh, *di);
+
+	return 0;
+}
+
+int ploop_create_dd(const char *ddxml, struct ploop_create_param *param)
+{
+	int ret;
+	struct ploop_disk_images_data *di = NULL;
+
+	ret = init_dd(&di, ddxml, param);
+	if (ret)
+		return ret;
+
+	ret = ploop_di_add_image(di, param->image, TOPDELTA_UUID, NONE_UUID);
+	if (ret)
+		goto err;
+
+	ret = ploop_store_diskdescriptor(ddxml, di);
+	if (ret)
+		goto err;
+
+err:
+	ploop_free_diskdescriptor(di);
+
+	return ret;
+}
+
+int ploop_create_image(struct ploop_create_param *param)
+{
+	struct ploop_disk_images_data *di = NULL;
+	char ddxml[PATH_MAX];
+	char fname[PATH_MAX];
+	int ret;
+	int fmt_version;
+
+	if (param->image == NULL) {
+		ploop_err(0, "Image file name not specified");
+		return SYSEXIT_PARAM;
+	}
+
+	get_disk_descriptor_fname_by_image(param->image, ddxml, sizeof(ddxml));
+	ret = init_dd(&di, ddxml, param);
+	if (ret)
+		return ret;
+
+	fmt_version = param->fmt_version == PLOOP_FMT_UNDEFINED ?
+		default_fmt_version() : param->fmt_version;
+
+	ret = create_image(param->image, di->blocksize, di->size,
+			param->mode, fmt_version);
 	if (ret)
 		goto out;
 
+	if (realpath(param->image, fname) == NULL) {
+		ploop_err(errno, "failed realpath(%s)", param->image);
+		ret = SYSEXIT_CREAT;
+		goto out;
+	}
+
+	ret = ploop_di_add_image(di, fname, TOPDELTA_UUID, NONE_UUID);
+	if (ret)
+		goto out;
+
+	ret = ploop_store_diskdescriptor(ddxml, di);
+	if (ret)
+		goto out;
+
+
 	if (param->fstype != NULL) {
 		ret = ploop_init_image(di, param);
 		if (ret)
-			ploop_drop_image(di);
+			goto out;
 	}
 
 out:
+	if (ret) {
+		if (di)
+			ploop_drop_image(di);
+		unlink(param->image);
+		unlink(ddxml);
+	}
+
 	ploop_free_diskdescriptor(di);
 
 	return ret;
@@ -912,6 +998,11 @@ static int ploop_stop(int fd, const char *devname)
 	return 0;
 }
 
+/* Returns:
+ *  0 mount point is found and saved to *out
+ *  1 mount point not found (fs not mounted)
+ * -1 some system error
+ */
 static int get_mount_dir(const char *device, char *out, int size)
 {
 	FILE *fp;
@@ -1718,6 +1809,23 @@ static int ploop_stop_device(const char *device)
 	return ret;
 }
 
+static int ploop_umount_fs(const char *mnt, struct ploop_disk_images_data *di)
+{
+	int ret;
+
+	/* The component_name feature allows multiple image mount.
+	 * Skip store statfs in custom case.
+	 */
+	if (di != NULL && di->runtime->component_name == NULL)
+		store_statfs_info(mnt, di->images[0]->file);
+	ploop_log(0, "Unmounting file system at %s", mnt);
+	ret = do_umount(mnt);
+	if (ret)
+		return ret;
+
+	return 0;
+}
+
 int ploop_umount(const char *device, struct ploop_disk_images_data *di)
 {
 	int ret;
@@ -1729,13 +1837,7 @@ int ploop_umount(const char *device, struct ploop_disk_images_data *di)
 	}
 
 	if (get_mount_dir(device, mnt, sizeof(mnt)) == 0) {
-		/* The component_name feature allows multiple image mount.
-		 * Skip store statfs in custom case.
-		 */
-		if (di != NULL && di->runtime->component_name == NULL)
-			store_statfs_info(mnt, di->images[0]->file);
-		ploop_log(0, "Unmounting file system at %s", mnt);
-		ret = do_umount(mnt);
+		ret = ploop_umount_fs(mnt, di);
 		if (ret)
 			return ret;
 	}
@@ -1794,7 +1896,6 @@ int get_image_param_online(const char *device, off_t *size,
 int get_image_param_offline(struct ploop_disk_images_data *di, const char *guid,
 		off_t *size, __u32 *blocksize, int *version)
 {
-	int ret;
 	struct delta delta;
 	const char *image;
 	int raw = 0;
@@ -1829,9 +1930,9 @@ int get_image_param_offline(struct ploop_disk_images_data *di, const char *guid,
 		*version = PLOOP_FMT_UNDEFINED;
 		*blocksize = di->blocksize;
 	} else {
-		ret = open_delta(&delta, image, O_RDONLY, OD_OFFLINE);
-		if (ret)
-			return ret;
+		if (open_delta(&delta, image, O_RDONLY, OD_OFFLINE))
+			return SYSEXIT_OPEN;
+
 		*size = delta.l2_size * delta.blocksize;
 		*version = delta.version;
 		*blocksize = delta.blocksize;
@@ -2067,6 +2168,7 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 	char buf[PATH_MAX];
 	char part_device[64];
 	int mounted = -1;
+	int umount_fs = 0;
 	int balloonfd = -1;
 	struct stat st;
 	off_t part_dev_size = 0;
@@ -2098,12 +2200,24 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 			goto err;
 
 		strncpy(mount_param.device, buf, sizeof(mount_param.device));
-		if (get_mount_dir(mount_param.device, buf, sizeof(buf))) {
-			ploop_err(0, "Can't find mount point for %s", buf);
-			ret = SYSEXIT_PARAM;
+		ret = get_mount_dir(mount_param.device, buf, sizeof(buf));
+		if (ret < 0) {
+			/* error message is printed by get_mount_dir() */
+			ret = SYSEXIT_SYS;
 			goto err;
-		}
-		mount_param.target = strdup(buf);
+		} else if (ret > 0) { /* not mounted */
+			ret = get_temp_mountpoint(di->images[0]->file, 1, buf, sizeof(buf));
+			if (ret)
+				goto err;
+
+			mount_param.target = strdup(buf);
+			ret = ploop_mount_fs(&mount_param);
+			if (ret)
+				goto err;
+
+			umount_fs = 1;
+		} else
+			mount_param.target = strdup(buf);
 		mounted = 1;
 	}
 
@@ -2287,7 +2401,7 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 						balloonfd, new_balloon_size);
 				if (ret)
 					goto err;
-				tune_fs(mount_param.target, part_device, new_fs_size);
+				tune_fs(balloonfd, part_device, new_fs_size);
 			}
 
 			if (new_balloon_size == 0) {
@@ -2303,6 +2417,9 @@ err:
 		close(balloonfd);
 	if (mounted == 0)
 		ploop_umount(mount_param.device, di);
+	else if (umount_fs)
+		ploop_umount_fs(mount_param.target, di);
+
 	ploop_unlock_dd(di);
 	free_mount_param(&mount_param);
 
