@@ -48,7 +48,6 @@ static int do_delete_snapshot(struct ploop_disk_images_data *di, const char *gui
 {
 	int ret;
 	char conf[PATH_MAX];
-	char *fname = NULL;
 	int nelem = 0;
 	char dev[64];
 	int snap_id;
@@ -74,10 +73,48 @@ static int do_delete_snapshot(struct ploop_disk_images_data *di, const char *gui
 
 	nelem = ploop_get_child_count_by_uuid(di, guid);
 	if (nelem == 0) {
-		if (strcmp(di->snapshots[snap_id]->parent_guid, NONE_UUID) == 0) {
+		struct ploop_snapshot_data *snap =  di->snapshots[snap_id];
+
+		if (strcmp(snap->parent_guid, NONE_UUID) == 0) {
 			ploop_err(0, "Unable to delete base image");
 			return SYSEXIT_PARAM;
 		}
+
+		if (strcmp(di->top_guid, guid) == 0) {
+			int id = find_snapshot_by_guid(di, snap->parent_guid);
+			if (id == -1) {
+				ploop_err(0, "Can't find snapshot by uuid %s",
+						snap->parent_guid);
+				return SYSEXIT_PARAM;
+			}
+
+			if (di->snapshots[id]->temporary) {
+				ploop_err(0, "Unable to delete top delta,"
+						" parent snapshot is temporary");
+				return SYSEXIT_PARAM;
+			}
+		}
+
+		char *fname = find_image_by_guid(di, guid);
+		if (fname == NULL) {
+			ploop_err(0, "Unable to find image by uuid %s",
+					guid);
+			return SYSEXIT_PARAM;
+		}
+
+		char *parent_fname = find_image_by_guid(di, snap->parent_guid);
+		if (parent_fname == NULL) {
+			ploop_err(0, "Unable to find image by uuid %s",
+					snap->parent_guid);
+			return SYSEXIT_PARAM;
+		}
+
+		ret = check_snapshot_mount(di, snap->temporary, parent_fname,
+				fname, guid);
+		if (ret)
+			return ret;
+
+		fname = NULL;
 		/* snapshot is not active and last -> delete */
 		ret = ploop_di_remove_image(di, guid, 1, &fname);
 		if (ret)
@@ -101,14 +138,15 @@ static int do_delete_snapshot(struct ploop_disk_images_data *di, const char *gui
 				guid);
 	} else if (nelem == 1) {
 		ret = ploop_merge_snapshot_by_guid(di, guid, PLOOP_MERGE_WITH_CHILD);
-	} else {
-		/* There no functionality to merge snapshot with >1 child */
-		ret = SYSEXIT_PARAM;
-		ploop_err(0, "There are %d references on %s snapshot: operation not supported",
-				nelem, guid);
+	} else if (!di->snapshots[snap_id]->temporary) {
+		ploop_log(1, "Warning: Unable to delete snapshot %s as there are %d references"
+				" to it; marking it as temporary instead",
+				guid, nelem);
+		di->snapshots[snap_id]->temporary = 1;
+		get_disk_descriptor_fname(di, conf, sizeof(conf));
+		ret = ploop_store_diskdescriptor(conf, di);
 	}
 
-
 	return ret;
 }
 
@@ -120,6 +158,8 @@ int ploop_delete_snapshot(struct ploop_disk_images_data *di, const char *guid)
 		return SYSEXIT_LOCK;
 
 	ret = do_delete_snapshot(di, guid);
+	if (ret == 0)
+		merge_temporary_snapshots(di);
 
 	ploop_unlock_dd(di);
 
@@ -230,10 +270,6 @@ static int do_create_snapshot(struct ploop_disk_images_data *di,
 	if (is_old_snapshot_format(di))
 		return SYSEXIT_PARAM;
 
-	ret = merge_temporary_snapshots(di);
-	if (ret)
-		return ret;
-
 	ret = gen_uuid_pair(snap_guid, sizeof(snap_guid),
 			file_guid, sizeof(file_guid));
 	if (ret) {
@@ -262,6 +298,33 @@ static int do_create_snapshot(struct ploop_disk_images_data *di,
 		return SYSEXIT_PARAM;
 	}
 
+	ret = ploop_find_dev_by_dd(di, dev, sizeof(dev));
+	if (ret == -1)
+		return SYSEXIT_SYS;
+	else if (ret == 0) {
+		online = 1;
+	} else {
+		ret = get_image_param_offline(di, di->top_guid, &size, &blocksize, &version);
+		if (ret == SYSEXIT_OPEN && errno == EBUSY) {
+			/* repair top delta */
+			char *topdelta[] = {find_image_by_guid(di, di->top_guid), NULL};
+			blocksize = di->blocksize;
+
+			ret = check_deltas(di, topdelta, 0, &blocksize);
+			if (ret)
+				return ret;
+
+			ret = get_image_param_offline(di, di->top_guid, &size, &blocksize, &version);
+		}
+
+		if (ret)
+			return ret;
+	}
+
+	ret = merge_temporary_snapshots(di);
+	if (ret)
+		return ret;
+
 	snprintf(fname, sizeof(fname), "%s.%s",
 			di->images[0]->file, file_guid);
 	ploop_di_change_guid(di, di->top_guid, snap_guid);
@@ -272,12 +335,6 @@ static int do_create_snapshot(struct ploop_disk_images_data *di,
 	if (ret)
 		return ret;
 
-	ret = ploop_find_dev_by_dd(di, dev, sizeof(dev));
-	if (ret == -1)
-		return SYSEXIT_SYS;
-	else if (ret == 0)
-		online = 1;
-
 	get_disk_descriptor_fname(di, conf, sizeof(conf));
 	snprintf(conf_tmp, sizeof(conf_tmp), "%s.tmp", conf);
 	ret = ploop_store_diskdescriptor(conf_tmp, di);
@@ -286,10 +343,6 @@ static int do_create_snapshot(struct ploop_disk_images_data *di,
 
 	if (!online) {
 		// offline snapshot
-		ret = get_image_param_offline(di, snap_guid, &size, &blocksize, &version);
-		if (ret)
-			goto err;
-
 		fd = create_snapshot_delta(fname, blocksize, size, version);
 		if (fd < 0) {
 			ret = SYSEXIT_CREAT;
@@ -484,15 +537,19 @@ int merge_temporary_snapshots(struct ploop_disk_images_data *di)
 		const char *guid = di->snapshots[i]->guid;
 
 		if (di->snapshots[i]->temporary &&
-				!is_snapshot_in_use(di, guid)) {
+				!is_snapshot_in_use(di, guid))
+		{
+			int nsnapshots = di->nsnapshots;
 
 			ret = do_delete_snapshot(di, guid);
 			if (ret)
 				return ret;
 
 			/* di has modified, start from beginning */
-			i = 0;
-			continue;
+			if (nsnapshots != di->nsnapshots) {
+				i = 0;
+				continue;
+			}
 		}
 		i++;
 	}
@@ -500,6 +557,53 @@ int merge_temporary_snapshots(struct ploop_disk_images_data *di)
 	return 0;
 }
 
+static int reset_top_delta(struct ploop_disk_images_data *di,
+		struct ploop_snapshot_switch_param *param)
+{
+	int ret;
+	char *old_top_delta_fname = NULL;
+	char conf[PATH_MAX];
+	char conf_tmp[PATH_MAX];
+	const char *guid = param->guid;
+
+	ret = ploop_di_remove_image(di, di->top_guid, 0, &old_top_delta_fname);
+	if (ret)
+		return ret;
+
+	ploop_di_change_guid(di, guid, TOPDELTA_UUID);
+
+	get_disk_descriptor_fname(di, conf, sizeof(conf));
+	snprintf(conf_tmp, sizeof(conf_tmp), "%s.tmp", conf);
+	ret = ploop_store_diskdescriptor(conf_tmp, di);
+	if (ret)
+		goto err;
+
+	if (rename(conf_tmp, conf)) {
+		ploop_err(errno, "Can't rename %s %s",
+				conf_tmp, conf);
+		ret = SYSEXIT_RENAME;
+		goto err;
+	}
+
+	/* destroy precached info */
+	drop_statfs_info(di->images[0]->file);
+
+	if (old_top_delta_fname != NULL) {
+		ploop_log(0, "Removing %s", old_top_delta_fname);
+		if (unlink(old_top_delta_fname))
+			ploop_err(errno, "Can't unlink %s",
+					old_top_delta_fname);
+	}
+
+err:
+	if (ret && unlink(conf_tmp))
+		ploop_err(errno, "Can't unlink %s", conf_tmp);
+
+	free(old_top_delta_fname);
+
+	return ret;
+}
+
 int ploop_switch_snapshot_ex(struct ploop_disk_images_data *di,
 		struct ploop_snapshot_switch_param *param)
 {
@@ -552,6 +656,11 @@ int ploop_switch_snapshot_ex(struct ploop_disk_images_data *di,
 		goto err_cleanup1;
 	}
 
+	if (flags & PLOOP_SNAP_SKIP_TOPDELTA_CREATE) {
+		ret = reset_top_delta(di, param);
+		goto err_cleanup1;
+	}
+
 	// Read image param from snapshot we going to switch on
 	ret = get_image_param(di, guid, &size, &blocksize, &version);
 	if (ret)
@@ -581,11 +690,13 @@ int ploop_switch_snapshot_ex(struct ploop_disk_images_data *di,
 			goto err_cleanup1;
 	} else if (param->guid_old != NULL) {
 		if (!is_valid_guid(param->guid_old)) {
+			ret = SYSEXIT_PARAM;
 			ploop_err(0, "Incorrect guid %s", param->guid_old);
 			goto err_cleanup1;
 		}
 
 		if (find_snapshot_by_guid(di, param->guid_old) != -1) {
+			ret = SYSEXIT_PARAM;
 			ploop_err(0, "Incorrect guid_old %s: already exists",
 					param->guid_old);
 			goto err_cleanup1;
@@ -594,15 +705,11 @@ int ploop_switch_snapshot_ex(struct ploop_disk_images_data *di,
 		ploop_di_change_guid(di, di->top_guid, param->guid_old);
 	}
 
-	if (flags & PLOOP_SNAP_SKIP_TOPDELTA_CREATE) {
-		ploop_di_change_guid(di, guid, TOPDELTA_UUID);
-	} else {
-		snprintf(new_top_delta_fname, sizeof(new_top_delta_fname), "%s.%s",
-			di->images[0]->file, file_uuid);
-		ret = ploop_di_add_image(di, new_top_delta_fname, TOPDELTA_UUID, guid);
-		if (ret)
-			goto err_cleanup1;
-	}
+	snprintf(new_top_delta_fname, sizeof(new_top_delta_fname), "%s.%s",
+		di->images[0]->file, file_uuid);
+	ret = ploop_di_add_image(di, new_top_delta_fname, TOPDELTA_UUID, guid);
+	if (ret)
+		goto err_cleanup1;
 
 	get_disk_descriptor_fname(di, conf, sizeof(conf));
 	snprintf(conf_tmp, sizeof(conf_tmp), "%s.tmp", conf);
@@ -611,14 +718,12 @@ int ploop_switch_snapshot_ex(struct ploop_disk_images_data *di,
 		goto err_cleanup1;
 
 	// offline snapshot
-	if (!(flags & PLOOP_SNAP_SKIP_TOPDELTA_CREATE)) {
-		fd = create_snapshot_delta(new_top_delta_fname, blocksize, size, version);
-		if (fd == -1) {
-			ret = SYSEXIT_CREAT;
-			goto err_cleanup2;
-		}
-		close(fd);
+	fd = create_snapshot_delta(new_top_delta_fname, blocksize, size, version);
+	if (fd == -1) {
+		ret = SYSEXIT_CREAT;
+		goto err_cleanup2;
 	}
+	close(fd);
 
 	if (rename(conf_tmp, conf)) {
 		ploop_err(errno, "Can't rename %s %s",
